from .xml import SubElement
from pptx.dml.color import RGBColor as RGB
from pptx.enum.dml import MSO_LINE
from pptx.util import Pt


# La dernière bordure ajoutée est gardée (à l'ajout d'une bordure, l'ancienne
# est retirée si elle existe).
# border_color est la couleur de la bordure
# border_width est la largeur de la bordure
# border_style est le style de la bordure
# left, top, right, bot indiquent quelles bordures sont à afficher
def set_cell_borders(cell,
                     border_color=RGB(0, 0, 0),
                     border_width=Pt(1),
                     border_style=MSO_LINE.DASH,
                     left=False,
                     top=False,
                     right=False,
                     bot=False):
    tc = cell._tc
    tcPr = tc.get_or_add_tcPr()

    borders = [('L', left), ('R', right), ('T', top), ('B', bot)]
    for idx, (side, do) in enumerate(borders):
        if not do:
            continue
        tag = f"a:ln{side}"

        # On retire la bordure précédente (si elle existe).
        tcPr.remove_all(tag)

        # les bordures doivent être déclarées dans un certain ordre (L, R, T, B)
        # sans quoi elles ne sont pas affichées...
        before = [f"a:ln{s}" for (s, _) in borders[idx + 1:]]
        # les bordures doivent apparaitre dans le xml avant les style de fond de la cellule
        # sinon elles ne sont pas affichées...
        before += ['a:solidFill', 'a:noFill']

        ln = SubElement(tcPr,
                        tag,
                        before=before,
                        w=str(border_width),
                        cap='flat',
                        cmpd='sng',
                        algn='ctr')

        solidFill = SubElement(ln, 'a:solidFill')
        # couleur rgb:
        colorval = "".join(map(lambda v: f"{v:02X}", border_color))
        SubElement(solidFill, 'a:srgbClr', val=colorval)
        # couleur prédéfinie:
        # SubElement(solidFill, 'a:schemeClr', val="tx1")

        SubElement(ln, 'a:prstDash', val=MSO_LINE.to_xml(border_style))
        SubElement(ln, 'a:round')
        SubElement(ln, 'a:headEnd', type='none', w='med', len='med')
        SubElement(ln, 'a:tailEnd', type='none', w='med', len='med')


# left, top, right, bot sont les coordonnées des cellules aux bordures
# outer indique si l'on veut afficher les bordures extérieures
# inner indique si l'on veut afficher les bordures intérieures
def set_rect_borders(table,
                     left,
                     top,
                     right,
                     bot,
                     outer=False,
                     inner=False,
                     **kwargs):
    assert left >= 0
    assert top >= 0
    assert right < len(table.columns)
    assert bot < len(table.rows)
    assert left <= right
    assert top <= bot

    for x in range(left, right + 1):
        for y in range(top, bot + 1):
            topb = outer if y == top else inner
            botb = outer if y == bot else inner
            leftb = outer if x == left else inner
            rightb = outer if x == right else inner

            set_cell_borders(table.cell(y, x),
                             **kwargs,
                             left=leftb,
                             top=topb,
                             right=rightb,
                             bot=botb)

    if not outer:
        return

    # PowerPoint a un bug (?), pour les cases intérieures du tableau
    # il ne regarde que les bordures droite et bas, il faut donc aussi
    # ajouter les bordures sur les cases extérieures de la selection

    for y in range(top, bot + 1):
        # bordure droite de la colonne de gauche
        if left != 0:
            set_cell_borders(table.cell(y, left - 1), **kwargs, right=True)
        # bordure gauche de la colonne de droite
        if right != len(table.columns) - 1:
            set_cell_borders(table.cell(y, right + 1), **kwargs, left=True)

    for x in range(left, right + 1):
        # bordure inférieure de la ligne du haut
        if top != 0:
            set_cell_borders(table.cell(top - 1, x), **kwargs, bot=True)
        # bordure supérieure de la ligne du bas
        if bot != len(table.rows) - 1:
            set_cell_borders(table.cell(bot + 1, x), **kwargs, top=True)


# size est la taille de la police, en général en Pt ou Cm:
# https://python-pptx.readthedocs.io/en/latest/api/util.html
# font est le nom de la font à utiliser (chaine de caractères, sensible à la casse)
# bold indique si le texte doit être en gras
# italic indique si le texte doit être en italique
# font_rgb est la couleur du texte, de type RGBColor, construit soit à partir de 3 entiers
# soit à partir d'une chaine de caractères représentant 3 hexadécimaux:
# https://python-pptx.readthedocs.io/en/latest/api/dml.html#pptx.dml.color.RGBColor
# font_brightness est la luminosité de la police, il s'agit d'un flottant entre -1 et 1
# underline indique si le texte doit être souligné
# language indique la langue du texte:
# https://python-pptx.readthedocs.io/en/latest/api/enum/MsoLanguageId.html
def font_style(ft,
               size=None,
               font=None,
               bold=None,
               italic=None,
               font_rgb=None,
               font_brightness=None,
               underline=None,
               language=None):
    if size is not None:
        ft.size = size
    if font is not None:
        ft.name = font
    if bold is not None:
        ft.bold = bold
    if italic is not None:
        ft.italic = italic
    if font_rgb is not None:
        ft.color.rgb = font_rgb
    if font_brightness is not None:
        ft.color.brightness = font_brightness
    if underline is not None:
        ft.underline = underline
    if language is not None:
        ft.language_id = language


# halign indique l'alignement horizontal du texte:
# https://python-pptx.readthedocs.io/en/latest/api/enum/PpParagraphAlignment.html
def text_style(tf, halign=None, **kwargs):
    for paragraph in tf.paragraphs:
        font_style(paragraph.font, **kwargs)
        if halign is not None:
            paragraph.alignment = halign


# valign indique l'alignement vertical du texte
# https://python-pptx.readthedocs.io/en/latest/api/enum/MsoVerticalAnchor.html
# back_rgb est la couleur de fond de la cellule
# https://python-pptx.readthedocs.io/en/latest/api/dml.html#pptx.dml.color.RGBColor
# back_brightness est la luminosité du fond de la cellule, un float entre -1 et 1
# margin_top, margin_left, margin_bottom, margin_right sont les marges de la cellule
# https://python-pptx.readthedocs.io/en/latest/api/util.html
def cell_style(cell,
               valign=None,
               back_rgb=None,
               back_brightness=None,
               margin_bottom=None,
               margin_top=None,
               margin_left=None,
               margin_right=None,
               **kwargs):
    if margin_bottom is not None:
        cell.margin_bottom = margin_bottom
    if margin_top is not None:
        cell.margin_top = margin_top
    if margin_right is not None:
        cell.margin_right = margin_right
    if margin_left is not None:
        cell.margin_left = margin_left
    if valign is not None:
        cell.vertical_anchor = valign
    if back_brightness is not None:
        cell.fill.solid()
        cell.fill.fore_color.brightness = back_brightness
    if back_rgb is not None:
        if not back_rgb:
            cell.fill.background()
        else:
            cell.fill.solid()
            cell.fill.fore_color.rgb = back_rgb

    text_style(cell.text_frame, **kwargs)


def paragraph_append(paragraph, txt, **kwargs):
    run = paragraph.add_run()
    run.text = txt
    font_style(run.font, **kwargs)



import logging
import sys

from pptx.oxml.text import CT_RegularTextRun, CT_TextCharacterProperties
from pptx.oxml.text import CT_TextLineBreak, CT_TextParagraphProperties

from pptx.enum.dml import MSO_COLOR_TYPE, MSO_FILL


def color_copy(cfrom, cto):
    if cfrom.type is not None:
        if cfrom.type == MSO_COLOR_TYPE.RGB:
            cto.rgb = cfrom.rgb
        elif cfrom.type == MSO_COLOR_TYPE.SCHEME:
            cto.theme_color = cfrom.theme_color
        else:
            logging.warning("font_copy: type de couleur inconnu '%s'",
                            cfrom.type)
        cto.brightness = cfrom.brightness


def fill_copy(ffrom, fto):
    if ffrom.type is None:
        return

    if ffrom.type == MSO_FILL.BACKGROUND:
        fto.background()
    elif ffrom.type == MSO_FILL.GRADIENT:
        fto.gradient()
        fto.gradient_angle = ffrom.gradient_angle
        fto.gradient_stops = ffrom.gradient_stops
    elif ffrom.type == MSO_FILL.PATTERNED:
        fto.patterned()
        fto.pattern = ffrom.pattern
        color_copy(ffrom.back_color, fto.back_color)
    elif ffrom.type == MSO_FILL.PICTURE:
        fto.picture()
    elif ffrom.type == MSO_FILL.SOLID:
        fto.solid()
    elif ffrom.type == MSO_FILL.TEXTURED:
        fto.textured()
    else:
        logging.warning("font_copy: type de remplissage inconnu '%s'",
                        ffrom.type)
    color_copy(ffrom.fore_color, fto.fore_color)


def font_copy(ffrom, fto):
    fto.bold = ffrom.bold
    fto.italic = ffrom.italic
    fto.language_id = ffrom.language_id
    fto.name = ffrom.name
    fto.size = ffrom.size
    fto.underline = ffrom.underline
    fill_copy(ffrom.fill, fto.fill)

    # attention accéder au champ color met le fill type à SOLID
    # il ne faut donc le faire que si ffrom.type == SOLID
    # (sinon ffrom.type est changé à SOLID)
    if ffrom.fill.type == MSO_FILL.SOLID:
        color_copy(ffrom.color, fto.color)


def run_copy(rfrom, rto):
    font_copy(rfrom.font, rto.font)
    rto.text = rfrom.text


def paragraph_copy(pfrom, pto):
    pto.clear()

    run_idx = 0
    # on itère directement sur les éléments xml car
    # c'est la seule manière de récupérer les sauts de ligne
    for e in pfrom._element:
        if isinstance(e, CT_RegularTextRun):
            # du texte (un run)
            run = pto.add_run()
            run_copy(pfrom.runs[run_idx], run)
            run_idx += 1
        elif isinstance(e, CT_TextLineBreak):
            # un saut de ligne
            pto.add_line_break()
        elif not isinstance(
                e, (CT_TextParagraphProperties, CT_TextCharacterProperties)):
            # CT_TextParagraphProperties est au début de chaque paragraphe et
            # contient les information de style du paragraphe, déjà copiées plus bas
            # CT_TextCharacterProperties est au début des runs qui ont un style différent,
            # et contient les informations sur ce style, copiées plus haut par run_copy
            logging.warning("paragraph_copy: element non prévu de type '%s'",
                            type(e))

    # il faut les copier après avoir créé les runs sans quoi
    # le style n'est pas complètement gardé
    # (add_line_break casse le niveau ?)
    font_copy(pfrom.font, pto.font)
    pto.alignment = pfrom.alignment
    pto.line_spacing = pfrom.line_spacing
    pto.space_after = pfrom.space_after
    pto.space_before = pfrom.space_before
    pto.level = pfrom.level


def textframe_copy(tffrom, tfto):
    tfto.clear()

    while len(tfto.paragraphs) != len(tffrom.paragraphs):
        pt = tfto.add_paragraph()

    for pf, pt in zip(tffrom.paragraphs, tfto.paragraphs):
        paragraph_copy(pf, pt)

    tfto.margin_bottom = tffrom.margin_bottom
    tfto.margin_left = tffrom.margin_left
    tfto.margin_right = tffrom.margin_right
    tfto.margin_top = tffrom.margin_top
    tfto.vertical_anchor = tffrom.vertical_anchor
    tfto.word_wrap = tffrom.word_wrap



from pptx.oxml.xmlchemy import OxmlElement


# https://github.com/scanny/python-pptx/issues/71
# https://groups.google.com/g/python-pptx/c/UTkdemIZICw
def SubElement(parent, tagname, before=(), **kwargs):
    element = OxmlElement(tagname)
    element.attrib.update(kwargs)
    if before:
        parent.insert_element_before(element, *before)
    else:
        parent.append(element)
    return element



import logging
import re
import sys

from PIL import Image

from .presmanager import delete_run

CMD_P1_ = CMD_P2_ = '{'
CMD_S1_ = CMD_S2_ = '}'

REPLACE_CHARS_ = {'‘': '\'', '’': '\''}


def cmd_p():
    return f"{CMD_P1_}{CMD_P2_}"


def cmd_s():
    return f"{CMD_S1_}{CMD_S2_}"


def cmd_pattern():
    return re.compile(f"{cmd_p()}(.*?){cmd_s()}")


def img_pattern():
    return f"{cmd_p()}img:(.*?):(.*?){cmd_s()}"


# https://docs.python.org/3/library/functions.html#exec
def interp(cmd, **kwargs):
    if '_r' in kwargs and '_c' in kwargs:
        context = f"[Slide {kwargs['_idx']}, cell at row {kwargs['_r']} and col {kwargs['_c']}]"
    else:
        context = f"[Slide {kwargs['_idx']}]"

    logging.debug("%s Interpreting command '%s'", context, cmd)
    for p, t in REPLACE_CHARS_.items():
        cmd = cmd.replace(p, t)
    try:
        exec(f"_ret={cmd}", None, kwargs)
    except Exception as e:
        logging.warning("%s Erreur à l'execution de la commande:\n%s\n%s",
                        context, cmd, e)
        # relancer l'exception ou juste continuer ?
        # raise
        return ''
    return kwargs['_ret']


# https://docs.python.org/3/library/re.html
def run_replace(run, **kwargs):
    cmd_pat = cmd_pattern()
    # ne pas utiliser subn car on veut retirer le code avant de l'executer
    m = cmd_pat.search(run.text)
    while m is not None:
        # supprime le code, puis interprète et rajoute le remplacement
        # évite les problèmes quand le code va dupliquer la slide par exemple
        run.text = run.text[:m.start()] + run.text[m.end():]
        val = str(interp(m.group(1), **kwargs))
        run.text = run.text[:m.start()] + val + run.text[m.start():]

        m = cmd_pat.search(run.text)


def cell_replace(cell, **kwargs):
    textframe_replace(cell.text_frame, **kwargs, _cell=cell)


def paragraph_replace(par, _p=None, **kwargs):
    for run in par.runs:
        run_replace(run, **kwargs, _p=par)

    idx = 0
    runs = list(par.runs)
    matches = list(cmd_pattern().finditer(par.text))
    for match in matches:
        idcs = -1  # char start
        while idcs == -1:
            rt = runs[idx].text
            if not rt:
                continue
            idcs = rt.find(cmd_p())

            if idcs == -1 and rt[-1] == CMD_P1_ and \
               runs[idx + 1].text[0] == CMD_P2_:
                idcs = len(rt) - 1

            idx += 1

        idrs = idx - 1  # run start

        idce = 1  # char end
        while idce == 1:
            rt = runs[idx].text
            if not rt:
                continue
            idce = rt.find(cmd_s()) + 2

            if idce == 1 and rt[-1] == CMD_S1_ and \
               runs[idx + 1].text[0] == CMD_S2_:
                idx += 2
                idce = 1
                break

            idx += 1

        idre = idx - 1  # run end

        runs[idrs].text = runs[idrs].text[:idcs]
        runs[idre].text = runs[idre].text[idce:]

        idx = idrs + 1

        for i in range(idrs + 1, idre):
            delete_run(runs[i])
        if not runs[idre].text:
            delete_run(runs[idre])

        # supprime le code, puis interprète et rajoute le remplacement
        # évite les problèmes quand le code va dupliquer la slide par exemple
        val = interp(match.group(1), **kwargs, _p=par)
        runs[idrs].text += str(val)

        if not runs[idrs].text:
            delete_run(runs[idrs])
        runs = list(par.runs)


def textframe_replace(tf, _tf=None, **kwargs):
    # match image
    m = re.search(img_pattern(), tf.text)

    if m is None:
        for par in tf.paragraphs:
            paragraph_replace(par, **kwargs, _tf=tf)
        return
        # don't replace accross paragraphs

    # on insère une image
    slide = tf.part.slide
    sh = tf._element.getparent()
    l, t, w, h = sh.x, sh.y, sh.cx, sh.cy

    modifs = set(m.group(1))
    file = interp(m.group(2), **kwargs, _tf=tf)
    slide.shapes.element.remove(sh)

    if not file:
        # ignore les placeholders avec image non définie
        return

    context = f"[Slide {kwargs['_idx']}] textframe_replace (image '{file}', modificateurs={modifs})"

    try:
        with Image.open(file, mode='r') as im:
            width = im.width
            height = im.height
    except (FileNotFoundError, PermissionError, OSError):
        logging.exception("%s: ", context)
        return

    hm = {'r', 'l', 'c'}
    vm = {'t', 'b', 'm'}
    allm = hm | vm

    if not modifs.issubset(allm):
        logging.info("%s: modificateurs inconnus %s ignorés", context,
                     modifs - allm)

    if len(modifs.intersection(hm)) > 1:
        logging.warning("%s: plusieurs modificateurs horizontaux %s", context,
                        modifs.intersection(hm))
        return
    if len(modifs.intersection(vm)) > 1:
        logging.warning("%s: plusieurs modificateurs verticaux %s", context,
                        modifs.intersection(vm))
        return

    factor = min(w / width, h / height)
    if modifs.isdisjoint(hm):
        width = w
    else:
        # scaling horizontal
        width *= factor
    if modifs.isdisjoint(vm):
        height = h
    else:
        # scaling vertical
        height *= factor

    if 'r' in modifs:
        # aligné à droite
        l += w - width
    if 'c' in modifs:
        # centré horizontalement
        l += (w - width) // 2
    if 'b' in modifs:
        # aligné en bas
        t += h - height
    if 'm' in modifs:
        # centré verticalement
        t += (h - height) // 2

    slide.shapes.add_picture(file, l, t, width=width, height=height)


def row_replace(row, _c=None, **kwargs):
    for _c in range(len(row.cells)):
        cell_replace(row.cells[_c], **kwargs, _c=_c)


def table_replace(table, _c=None, _r=None, **kwargs):
    for _r in range(len(table.rows)):
        row_replace(table.rows[_r], **kwargs, _r=_r)


def slide_replace(slide, _sl=None, **kwargs):
    for sh in slide.shapes:
        if sh.has_text_frame:
            textframe_replace(sh.text_frame, **kwargs, _sl=slide)
        if sh.has_table:
            table_replace(sh.table, **kwargs, _sl=slide)


def pres_replace(pres, _idx=None, _pres=None, **kwargs):
    for idx, slide in enumerate(pres.slides):
        slide_replace(slide, **kwargs, _pres=pres, _idx=idx)



import logging
import os
import sys

import copy

from pptx import Presentation
from pptx.chart.data import CategoryChartData
from pptx.table import _Column, _Row

from .copy import textframe_copy


def export(pres, outpath):
    try:
        os.remove(outpath)
        logging.info(f"Former {outpath} deleted.")
    except OSError:
        pass
    pres.save(outpath)


def remove_slide(pres, idx):
    rId = pres.slides._sldIdLst[idx].rId
    pres.part.drop_rel(rId)
    del pres.slides._sldIdLst[idx]


def delete_run(r):
    r._r.getparent().remove(r._r)


def clear_slides(pres):
    for idx in range(len(pres.slides) - 1, -1, -1):
        remove_slide(pres, idx)


def insert_slide(pres, layout, idx):
    slide = append_slide(pres, layout)
    move_slide(pres.slides, slide, idx)
    return slide


def move_slide(slides, slide, new_idx):
    slides._sldIdLst.insert(new_idx, slides._sldIdLst[slides.index(slide)])


def fill_placeholders(slide):
    layout = slide.slide_layout
    for ph in layout.placeholders:
        if ph.has_text_frame:
            try:
                phid = ph.placeholder_format.idx
                tf = slide.placeholders[phid].text_frame
            except KeyError:
                continue
            textframe_copy(ph.text_frame, tf)


# Ajoute la slide avec le nom donné à la fin de la présentation
def append_slide(pres, layout):
    if isinstance(layout, str):
        nl = pres.slide_masters[0].slide_layouts.get_by_name(layout)
        if nl is None:
            logging.error("append_slide: layout \'%s\' inexistant", layout)
            sys.exit(1)
        layout = nl

    slide = pres.slides.add_slide(layout)
    fill_placeholders(slide)
    return slide


# Ecrit le numéro des shapes, pratique pour repérer initialement
def etalonnage(prespath, outpath):
    pres = Presentation(prespath)
    for idx, slide in enumerate(pres.slides):
        logging.info("Slide %d has slide_id %s", idx, slide.slide_id)

    clear_slides(pres)
    for layout in pres.slide_masters[0].slide_layouts:
        slide = pres.slides.add_slide(layout)
        for ph in slide.placeholders:
            ph.text = f"placeholder {ph.placeholder_format.idx}"
    export(pres, outpath)


def chart_replace_data(chart, values):
    chart_data = CategoryChartData()
    chart_data.categories = chart.plots[0].categories
    chart_data.add_series('Values', values)
    chart.replace_data(chart_data)


def find_shape(slide, name):
    for shape in slide.shapes:
        if shape.name == name:
            return shape
    return None


# duplique la dernière ligne, en particulier les bordures et le texte
# sont gardés exactement à l'identique
def table_dup_row(table, idx, n=1, to=None):
    if to is None:
        to = idx
    to = (to + len(table.rows)) % len(table.rows)

    new_row = table._tbl.tr_lst[idx]
    num = -1
    for rid, rxml in enumerate(table._tbl.iterchildren()):
        if rxml.tag == new_row.tag:
            num += 1
            if num == to:
                break

    for _ in range(n):
        new_row = copy.deepcopy(new_row)
        table._tbl.insert(rid, new_row)

    if n == 1:
        return _Row(new_row, table)
    return None


def table_dup_column(table, idx, n=1, to=None, keep_width=False):
    total_width = table._graphic_frame.width
    colwidth = table.columns[idx].width
    mult = total_width / (total_width + n * colwidth)

    if to is None:
        to = idx
    to = (to + len(table.columns)) % len(table.columns)

    new_col = table._tbl.tblGrid.gridCol_lst[idx]
    num = -1
    for cid, cxml in enumerate(table._tbl.tblGrid.iterchildren()):
        if cxml.tag == new_col.tag:
            num += 1
            if num == to:
                break
    for _ in range(n):
        new_col = copy.deepcopy(new_col)
        table._tbl.tblGrid.insert(cid, new_col)

    for tr in table._tbl.tr_lst:
        new_cell = tr.tc_lst[idx]
        num = -1
        for cid, cxml in enumerate(tr.iterchildren()):
            if cxml.tag == new_cell.tag:
                num += 1
                if num == to:
                    break
        for _ in range(n):
            new_cell = copy.deepcopy(new_cell)
            tr.insert(cid, new_cell)

    if keep_width:
        for col in table.columns:
            col.width = int(col.width * mult)

    if n == 1:
        return _Column(new_col, table)
    return None


# def _get_blank_slide_layout(pres):
#     layout_items_count = [
#         len(layout.placeholders) for layout in pres.slide_layouts
#     ]
#     min_items = min(layout_items_count)
#     blank_layout_id = layout_items_count.index(min_items)
#     return pres.slide_layouts[blank_layout_id]


def duplicate_slide(pres, index):
    source = pres.slides[index]
    dest = pres.slides.add_slide(source.slide_layout)

    for shp in source.shapes:
        el = shp.element
        newel = copy.deepcopy(el)
        dest.shapes._spTree.insert_element_before(newel, 'p:extLst')

    for key, value in source.rels.items():
        if not "notesSlide" in value.reltype:
            dest.rels.add_relationship(value.reltype, value._target, value.rId)

    move_slide(pres.slides, pres.slides[-1], index)

    return dest
